@@ -5,7 +5,8 @@ use strict;
 use warnings;
 use Socket;
 use IO::Socket::SSL;
-use IO::Lambda qw(:lambda :stream);
+use IO::Lambda qw(:lambda :stream :dev :constants);
+use Errno qw(EWOULDBLOCK EAGAIN);
 
 our $DEBUG = $IO::Lambda::DEBUG{https};
 
@@ -64,10 +65,51 @@ sub https_connect
 	}
 }
 
+sub https_syscall
+{
+	my ( $read, $fh, $buf, $length, $offset) = @_;
+	$$buf = '' unless defined $$buf;
+	local $SIG{PIPE} = 'IGNORE';
+	my $n = $read ? 
+		sysread( $fh, $$buf, $length, $offset) : 
+		syswrite( $fh, $$buf, $length, $offset);
+	unless ( defined $n ) {
+		my $err = $!;
+		warn "fh(", fileno($fh), ") ", ( $read ? 'read' : 'write'), " error $err\n" if $DEBUG;
+		return undef, $err;
+	}
+	if ( $DEBUG ) {
+		warn "fh(", fileno($fh), ") ", ( $read ? 'read' : 'wrote'), "$n bytes\n";
+		warn substr( $$buf, length($$buf) - $n), "\n" if $DEBUG > 1 and $n > 0;
+	}
+	return $n;
+}
+
+sub https_syscall_watcher
+{
+	my $read = shift;
+	lambda {
+		my ( $fh, $buf, $length, $offset, $deadline) = @_;
+
+		($deadline, $offset) = ($offset, length($$buf) || 0) if $read;
+
+		my ( $n, $err ) = https_syscall( $read, $fh, $buf, $length, $offset );
+		return ($n, $err) if defined($n) || ($err != EAGAIN && $err != EWOULDBLOCK);
+
+		this-> watch_io( $read ? IO_READ : IO_WRITE, $fh, $deadline, _subname https_syscall_watcher => sub {
+			return undef, 'timeout' unless $_[1];
+			my ( $n, $err ) = https_syscall( $read, $fh, $buf, $length, $offset );
+			return $n if defined $n;
+			$err = $SSL_ERROR if $err == EWOULDBLOCK || $err == EAGAIN;
+			return undef, $err;
+		});
+	};
+}
+
 sub https_writer
 {
 	my $cached = shift;
-	my $writer = syswriter;
+	my $writer = https_syscall_watcher(0);
 
 	lambda {
 		my ( $sock, $req, $length, $offset, $deadline) = @_;
@@ -87,7 +129,7 @@ sub https_writer
 
 sub https_reader
 {
-	my $reader = sysreader;
+	my $reader = https_syscall_watcher(1);
 	lambda {
 		my ( $sock, $buf, $length, $deadline) = @_;
 		context $reader, $sock, $buf, $length, $deadline;
